/*
 * TelPac-Node/LinuX developed by Hans-J. Barthen, DL5DI based on tcp_call
 * from ax25-tools 0.0.8, enhanced for full-binary transfer and automatic
 * password protected user- and node-logon at Winlink 2000 servers.
 *
 * This code is like the original tools term of the GNU licence.
 *
 * In case of problems please contact dl5di@darc.de/dl5di@db0lj.ampr.org
 * Hans-Juergen Barthen, Franz-Josef-Str. 20, D-56642 Kruft
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>

#include "../user_io.h"
#include "../tpcfg.h"

#define BUFLEN	8192

int paclen = 128;  // set to "safe" default value

/* This is for select_loop() */
static unsigned char buf[BUFLEN];

#define min(a,b)	((a) < (b) ? (a) : (b))

void err(char *message)
{
    FILE *err_file;
    FILE *log_file;
    time_t tstop;
	send(STDOUT_FILENO, message, strlen(message), 0);
	time(&tstop);

	if((err_file = fopen(errfile, "a")) != NULL) {
	    fprintf(err_file, "Logout '%-24.24s' %s ERROR !\n",message, ctime(&tstop));
	    fclose(err_file);
	}

	if((log_file = fopen(logfile, "a")) != NULL) {
	    fprintf(log_file, "Logout %-24.24s ERROR !\n",ctime(&tstop));
	    fclose(log_file);
	}

	exit(1);
}

void msg(char *message)
{
    FILE *err_file;
    time_t tstop;

	if((err_file = fopen(errfile, "a")) != NULL) {
	    time(&tstop);
	    fprintf(err_file, "%s %s", message, ctime(&tstop));
	    fclose(err_file);
	}

	return;
}

struct ust select_loop(int s, char *usercall, char *passwd, char *nodecall)
{
	fd_set read_fd;
	int ffup = 0; 
	int ffdn = 0; 
	int m = 0;
	int n = 0;
	int init = 0;
	int loop = 0;
	unsigned char *cp;
	struct ust rc;
        char mybuf[80];
        int res;
	
	rc.errcode = 0;
	rc.bytes_recv = 0;
	rc.bytes_sent = 0;

	if (fcntl(s, F_SETFL, O_NONBLOCK) < 0) {
	        // probably should change this to preserve any
	        // other flags that may be set  E.G.
	        //   arg = fcntl(mysock, F_GETFL, NULL); 
  		//   arg |= O_NONBLOCK; 
		//    fcntl(mysock, F_SETFL, arg);   
		close(s);
		rc.errcode = -1;	/* can not open socket */
		return(rc);
	}
	if (fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK) < 0) {
		close(s);
		rc.errcode = -2;	/* can not open ax25 */
		return(rc);
	}

	/*
	 * Loop until one end of the connection goes away.
	 */

	for (;;) {
		FD_ZERO(&read_fd);
		FD_SET(STDIN_FILENO, &read_fd);
		FD_SET(s, &read_fd);
 
		res = select(s + 1, &read_fd, NULL, NULL, NULL);

                //sprintf(mybuf,"Info: Res = %d, s = %d\r\n", res, s);
                //send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);
                
                if (FD_ISSET(s, &read_fd))
                // If the socket indicated ready to read but returns 0 bytes
                // on a receive then the other end has closed.
                // Only ask for one byte to minimize work done by the following
                // recv statement as it is called every time through the loop.
                  {
                  //n = recv(s, buf, 1, MSG_PEEK | MSG_DONTWAIT);
                  if ((n = recv(s, buf, 1, MSG_PEEK | MSG_DONTWAIT)) == 0)
                    {
                    sprintf(mybuf,"; INFO: Connection closed by PMBO\r\n");
                    send(STDOUT_FILENO, mybuf, strlen(mybuf), 0);                                
                    rc.errcode = 0;
                    sleep(1);
                    close(s);
                    return(rc);
                    }
                    
                  } //if (FD_ISSET(s, &read_fd))
                 
/*
 *	node logon to server after recept of 1st buffer
 */
 
		if ((init == 0) && (FD_ISSET(s, &read_fd))) {
			while ((n = recv(s, buf, BUFLEN, 0)) > 0) {		
			    /* read it and eat it ... */
			    /*  discard server ctext  */
			    rc.bytes_recv += n;
			    m = n;
			}
			
			if(*(buf+m-2) == ':'){

			    n = sprintf(buf,".%s\r%s\rTelPac %s\r",usercall,passwd,nodecall);
			    rc.bytes_sent += n;
			    cp = buf;
			    loop = 0;

    			    while (n > 0) {
				m = send(s, cp, n, 0);
				if (m > 0) {
			    	    cp += m;
			    	    n -= m;
				}
			    }

			    if(n < 0 && errno != EAGAIN) {
				close(s);
				rc.errcode = -3;
				break;
			    }

			    init = 1;
			}
		}

/*
 *	Receive from WL2K-Server and send to AX.25
 *	analyze last 4 characters (EOL on telnet link is CR/LF)
 */
		if (FD_ISSET(s, &read_fd)) {
			while ((n = recv(s, buf, BUFLEN, 0)) > 0) {
			    rc.bytes_sent += n;
			    cp = buf;
			    loop = 0;
			    // Uplink sent FF ?
			    if ((*(cp+n-3) == 'F') && (*(cp+n-2) == 'F')) {
			        ffup = 1;
			    } else {
			    // Uplink sent FQ after Downlink sent FF ?
			        if ((n == 3) && (ffdn == 1)) {
				    if ((*cp == 'F') && (*(cp+1) == 'Q')) {
					loop = TIMEOUT;
					rc.errcode = 10; 	//normal final
				    } else {
				        ffdn = 0;
				    }
				}
			    }
                            
    			    while (n > 0) {	/* send buffer transparent on ax25 */
                                m = send(STDOUT_FILENO, cp, min(paclen, n), 0);
				if (m > 0) {
				    cp += m;
				    n -= m;
				}
			    }
			    if(n < 0 && errno != EAGAIN) {
				close(s);
				rc.errcode = -4;
				break;
			    }
			}
		}
			
/*
 *	Receive from AX.25 and send to WL2K-Server
 *	analyze last 3 characters (EOL on ax25 link is CR)
 */
		if (FD_ISSET(STDIN_FILENO, &read_fd)) {
			while ((n = recv(STDIN_FILENO, buf, BUFLEN, 0)) > 0) {
			    rc.bytes_recv += n;
			    cp = buf;
			    loop = 0;
			    // Downlink sent FF ?
			    if ((*(cp+n-3) == 'F') && (*(cp+n-2) == 'F')) {
			        ffdn = 1;
			    } else {
			    // Downlink sent FQ after Uplink sent FF ?
			        if ((n == 3) && (ffup == 1)) {
				    if ((*cp == 'F') && (*(cp+1) == 'Q')) {
					loop = TIMEOUT;
					rc.errcode = 10; 	//normal final
				    } else {
				        ffup = 0;
				    }
				}
			    }

			    while (n > 0) {	/* send buffer transparent on telnet */
				m = send(s, cp, n, 0);
				if (m > 0) {
				    cp += m;
				    n -= m;
				}
			    }
			    if(n < 0 && errno != EAGAIN) {
				close(s);
				rc.errcode = -5;
				break;
			    }
			}
		}
		
		if(loop > 0){
		    sleep(1);
		}

		loop++;
		
		if(loop > TIMEOUT) {
		    if (rc.errcode == 10){
			rc.errcode = 0;		// normal final
		    } else {
			rc.errcode = 1;		// TimeOut
		    }
		    close(s);
		    return(rc);
		}
	}
	return(rc);
}


/*
 *  write logintime and usercall to logfile
 */

int log_logon(char *usercall, time_t tstart,char *hostname)
{
    FILE *log_file;

    if((log_file = fopen(logfile, "a")) != NULL) {
	fprintf(log_file, "Login  %-9.9s %-24.24s connected to %s\n",usercall,ctime(&tstart),hostname);
	fclose(log_file);
    }
    return (0);
}


/*
 *  write logouttime, usercall and statistics to logfile
 */


int log_logout(char *usercall, struct ust userstat, time_t tstart, time_t tstop)
{
    double tused;
    float rate;
    FILE *log_file;

    if((log_file = fopen(logfile, "a")) != NULL) {
	time(&tstop);
	tused = difftime(tstop, tstart);
        rate = (userstat.bytes_sent + userstat.bytes_recv) / tused;
	fprintf(log_file, "Logout %-9.9s %-24.24s tx:%li rx:%li %.1fs %.1f Bytes/s (%d)\n",usercall,ctime(&tstop),userstat.bytes_sent,userstat.bytes_recv,tused,rate,userstat.errcode);
	fclose(log_file);
    }
    return (0);
}
